function [Graphs, a_n] = sampleDags(nsamples, nnodes)
% Samples 'nsamples' DAGs with 'nnodes' nodes uniformly at random
% Returns the DAGs in the struct 'Graphs' and the number of DAGs with
% 'nnodes' nodes a_n
    function val = computeAlpha(n,k)
        if(~isnan(a_nk(n,k)))
            val = a_nk(n,k);
        else
            val = nchoosek(n,k) * computeBeta(n,k);
            a_nk(n,k) = val;
        end
    end

    function val = computeAlpha_sym(n,k)
        if(~isnan(a_nk(n,k)))
            val = a_nk(n,k);
        else
            [n,k]
            val = nchoosek(vpa(sym(n)),vpa(sym(k))) * computeBeta_sym(n,k);
            a_nk(n,k) = val;
        end
    end

    function val = computeBeta(n,k)
        if(~isnan(b_nk(n,k)))
            val = b_nk(n,k);
        else
            val = 0;
            m = n - k;
            for s = 1:m
                val = val + (2^k-1)^s*2^(k*(m-s))*computeAlpha(m,s);
            end
            b_nk(n,k) = val;
        end
    end

    function val = computeBeta_sym(n,k)
        if(~isnan(b_nk(n,k)))
            val = b_nk(n,k);
        else
            val = vpa(sym(0));
            m = n - k;
            for s = 1:m
                val = val + (2^vpa(sym(k))-1)^s*2^(vpa(sym(k))*(m-s))*computeAlpha_sym(m,s);
            end
            b_nk(n,k) = val;
        end
    end

    function kappa = computeKappa(n,k)
        if(n == k)
            kappa = [];
            return;
        end
        
        m = n-k;
        
        ndigits = ceil(log2(b_nk(n,k)));
        if(ndigits <= 52)
            r = randi(b_nk(n,k));
        else
            r = sampleBinary(ndigits,pow);
            while(r > b_nk(n,k))
                r = sampleBinary(ndigits,pow);
            end
        end
        s = m;
        cursum = 0;
        for i = 1:m
            cursum = cursum + (2^k-1)^i*2^(k*(m-i))*a_nk(m,i);
            if(cursum >= r)
                s = i;
                break;
            end
        end
        kappa = [s computeKappa(m,s)];
    end

    function kappa = computeKappa_sym(n,k)
        if(n == k)
            kappa = [];
            return;
        end
        
        m = n-k;
        
        ndigits = double(ceil(log2(b_nk(n,k))));
        if(ndigits <= 52)
            r = randi(double(b_nk(n,k)));
        else
            r = sampleBinary(ndigits,pow);
            while(r > b_nk(n,k))
                r = sampleBinary(ndigits,pow);
            end
        end
        s = m;
        cursum = 0;
        for i = 1:m
            cursum = cursum + (2^vpa(sym(k))-1)^i*2^(vpa(sym(k))*(m-i))*a_nk(m,i);
            if(cursum >= r)
                s = i;
                break;
            end
        end
        kappa = [s computeKappa_sym(m,s)];
    end

    function G = sampleDag
        if(maxdigits <= 52)
            r = randi(a_n);
        else
            r = sampleBinary(maxdigits,pow);
            while(r > a_n)
                r = sampleBinary(maxdigits,pow);
            end
        end
        k_ = find(a_cumsum >= r,1);
        kappa = [k_ computeKappa(nnodes,k_)];
        kappa = fliplr(kappa);
        
        % Now create G
        tmpG = rand(nnodes,nnodes) < 0.5;
        for i = 1:nnodes
            tmpG(i,i) = 0;
        end
        
        k_cumsum = [0 cumsum(kappa)];
        
        % First sample all
        tmpG(1:kappa(1),1:kappa(1)) = 0;
        for i = 1:length(kappa)-1
            tmpG(1:k_cumsum(i+2),k_cumsum(i+1)+1:end) = 0;
        end
        
        % Now check for zero columns
        for i = 1:length(kappa)-1
            r_from = k_cumsum(i+1)+1;
            r_to = k_cumsum(i+2);
            col_from = k_cumsum(i)+1;
            col_to = k_cumsum(i+1);
            
            for c = col_from:col_to
                while(~any(tmpG(r_from:r_to,c)))
                    tmpG(r_from:r_to,c) = rand(r_to-r_from+1,1) < 0.5;
                end
            end
        end
        
        % Now random permutation
        perm = randperm(nnodes);
        G = tmpG(perm,perm);
    end

    function G = sampleDag_sym
        if(maxdigits <= 52)
            r = randi(double(a_n));
        else
            r = sampleBinary(maxdigits,pow);
            while(r > a_n)
                r = sampleBinary(maxdigits,pow);
            end
        end
        k_ = find(a_cumsum >= r,1);
        kappa = [k_ computeKappa_sym(nnodes,k_)];
        kappa = fliplr(kappa);
        
        % Now create G
        tmpG = rand(nnodes,nnodes) < 0.5;
        for i = 1:nnodes
            tmpG(i,i) = 0;
        end
        
        k_cumsum = [0 cumsum(kappa)];
        
        % First sample all
        tmpG(1:kappa(1),1:kappa(1)) = 0;
        for i = 1:length(kappa)-1
            tmpG(1:k_cumsum(i+2),k_cumsum(i+1)+1:end) = 0;
        end
        
        % Now check for zero columns
        for i = 1:length(kappa)-1
            r_from = k_cumsum(i+1)+1;
            r_to = k_cumsum(i+2);
            col_from = k_cumsum(i)+1;
            col_to = k_cumsum(i+1);
            
            for c = col_from:col_to
                while(~any(tmpG(r_from:r_to,c)))
                    tmpG(r_from:r_to,c) = rand(r_to-r_from+1,1) < 0.5;
                end
            end
        end
        
        % Now random permutation
        perm = randperm(nnodes);
        G = tmpG(perm,perm);
    end

a_nk = zeros(nnodes,nnodes) + NaN;
b_nk = zeros(nnodes,nnodes) + NaN;

for n = 1:nnodes
    a_nk(n,n) = 1;
    b_nk(n,n) = 0;
end

for n = 1:nnodes
    for k = 1:n
        a_nk(n,k) = computeAlpha(n,k);
        b_nk(n,k) = computeBeta(n,k);
    end
end

a_n = sum(a_nk(end,:));

if(isinf(a_n))
    %     error('Can''t handle such large networks yet :(');
    a_nk = vpa(sym(zeros(nnodes,nnodes) + NaN));
    b_nk = vpa(sym(zeros(nnodes,nnodes) + NaN));
    
    for n = 1:nnodes
        a_nk(n,n) = 1;
        b_nk(n,n) = 0;
    end
    
    for n = 1:nnodes
        for k = 1:n
            a_nk(n,k) = computeAlpha_sym(n,k);
            b_nk(n,k) = computeBeta_sym(n,k);
        end
    end
    a_n = sum(a_nk(end,:));

    a_cumsum = vpa(sym(zeros(1,nnodes)));
    a_cumsum(1) = a_nk(end,1);
    for i_ = 2:nnodes
        a_cumsum(i_) = a_cumsum(i_-1) + a_nk(end,i_);
    end
    set(0,'RecursionLimit',10000);
    maxdigits = double(ceil(log2(a_n)));
    pow = 2.^(0:maxdigits-1);
    Graphs = cell(nsamples,1);
    
    for ctr = 1:nsamples
        Graphs{ctr} = double(sampleDag_sym);
    end
else
    a_cumsum = cumsum(a_nk(end,:));
    
    maxdigits = ceil(log2(a_n));
    pow = 2.^(0:maxdigits-1);
    
    Graphs = cell(nsamples,1);
    
    for ctr = 1:nsamples
        Graphs{ctr} = double(sampleDag);
        if(mod(ctr,1000) == 0)
            ctr
        end
    end
end

end
